/**
 * Fractal Binding Factory.
 * Copyright (C) 2007-2009 INRIA, SARDES

 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contact: Valerio.Schiavoni@gmail.com
 * 
 * Author: Valerio Schiavoni
 * 
 */
package org.objectweb.fractal.bf.connectors.rmi.asm;

import java.io.PrintWriter;
import java.lang.reflect.Method;
import java.util.logging.Logger;

import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.FieldVisitor;
import org.objectweb.asm.Label;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;
import org.objectweb.asm.util.CheckClassAdapter;
import org.objectweb.fractal.bf.BindingFactoryClassLoader;
import org.objectweb.fractal.bf.proxies.BytecodeVerifier;

/**
 * Generate a class that implements the generated subclass of
 * {@link java.rmi.Remote}.
 */
public class RemoteSubclassGenerator implements Opcodes {

	/**
	 * The name of the field to generate. This Java type of this field is
	 * derived by the java type of the interface passed as first argument to the
	 * {@link RemoteSubclassGenerator#generate(Class, Class, BindingFactoryClassLoader)}
	 * method. It is assigned to the business object in charge of answering to
	 * calls.
	 */
	private static final String DELEGATED_FIELD_NAME = "s";
	Logger log = Logger.getLogger(RemoteSubclassGenerator.class
			.getCanonicalName());

	/**
	 * Generate and define a class that can be exported into a RMI registry.
	 * 
	 * @param unremotableInterfaze
	 *            the type of the delegate field to which calls will be
	 *            delegated to
	 * @param remotableInterfaze
	 *            a subclass of java.rmi.Remote, that is implemented by the
	 *            resulting class
	 * @param cl
	 *            the {@link BindingFactoryClassLoader} used to to define the
	 *            class.
	 * @throws RuntimeException
	 *             if the generated bytecode doesn't pass the checks performed
	 *             by
	 *             {@link CheckClassAdapter#verify(ClassReader, boolean, PrintWriter)}
	 *             method.
	 * @see CheckClassAdapter
	 */
	public Class<?> generate(Class<?> unremotableInterfaze,
			Class<?> remotableInterfaze, BindingFactoryClassLoader cl) {
		final String unremotableInterfazeInternalName = Type
				.getInternalName(unremotableInterfaze);
		final String remotableInterfazeInternalName = Type
				.getInternalName(remotableInterfaze);

		log.finest("Internal type of delegate field: "
				+ unremotableInterfazeInternalName);
		/*
		 * Using COMPUTE_MAXS flag avoids to handle manually stack indices, to
		 * be specified in the various computeMaxs methods.
		 */
		ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS);

		FieldVisitor fv;
		MethodVisitor mv;

		final String generatedClassInternalName = remotableInterfazeInternalName
				+ "Impl";
		log.finest("Generated class internal name: "
				+ generatedClassInternalName);
		cw.visit(V1_5, ACC_PUBLIC + ACC_SUPER, generatedClassInternalName,
				null, "java/lang/Object",
				new String[] { remotableInterfazeInternalName,
						"java/io/Serializable" });

		{

			fv = cw.visitField(ACC_PUBLIC, DELEGATED_FIELD_NAME, "L"
					+ unremotableInterfazeInternalName + ";", null, null);
			fv.visitEnd();
			log.finest("Field name '" + DELEGATED_FIELD_NAME + "' of type "
					+ unremotableInterfazeInternalName + " generated");
		}
		/*
		 * Generate 1-arg constructor. The type of the parameter is the same as
		 * the unremotableInterfaze parameter.
		 */
		{
			mv = cw.visitMethod(ACC_PUBLIC, "<init>", "(L"
					+ unremotableInterfazeInternalName + ";)V", null,
					new String[] { "java/rmi/RemoteException" });
			mv.visitCode();
			Label l0 = new Label();
			mv.visitLabel(l0);
			mv.visitVarInsn(ALOAD, 0);
			mv.visitMethodInsn(INVOKESPECIAL, "java/lang/Object", "<init>",
					"()V");
			Label l1 = new Label();
			mv.visitLabel(l1);
			mv.visitVarInsn(ALOAD, 0);
			mv.visitVarInsn(ALOAD, 1);
			mv.visitFieldInsn(PUTFIELD, generatedClassInternalName,
					DELEGATED_FIELD_NAME, "L"
							+ unremotableInterfazeInternalName + ";");
			Label l2 = new Label();
			mv.visitLabel(l2);
			mv.visitInsn(RETURN);
			Label l3 = new Label();
			mv.visitLabel(l3);
			mv.visitLocalVariable("this", "L" + generatedClassInternalName
					+ ";", null, l0, l3, 0);
			mv.visitLocalVariable(DELEGATED_FIELD_NAME, "L"
					+ unremotableInterfazeInternalName + ";", null, l0, l3, 1);
			mv.visitMaxs(2, 2);
			mv.visitEnd();
			log.fine("1-Args constructor generated.");

		}
		/*
		 * for every method declared by the remotableInterfaze interface,
		 * generate a method in the class being generated with the same
		 * signature
		 */
		for (Method method : remotableInterfaze.getMethods()) {
			log.finest("Generating method body for " + method.getName());

			Class<?>[] paramTypes = method.getParameterTypes();
			int args = paramTypes.length;

			String desc = Type.getMethodDescriptor(method);

			mv = cw.visitMethod(ACC_PUBLIC, method.getName(), desc, null,
					new String[] { "java/rmi/RemoteException" });

			mv.visitCode();

			Label l0 = new Label();
			mv.visitLabel(l0);
			/*
			 * push 'this' on the stack
			 */
			mv.visitVarInsn(ALOAD, 0);
			/*
			 * recover a reference to the delegate field for 'this' object
			 */
			mv.visitFieldInsn(GETFIELD, generatedClassInternalName,
					DELEGATED_FIELD_NAME, "L"
							+ unremotableInterfazeInternalName + ";");
			Label l1 = new Label();
			mv.visitLabel(l1);
			mv.visitLocalVariable("this", "L" + generatedClassInternalName
					+ ";", null, l0, l1, 0);
			/*
			 * for every parameter of the current method, push the actual
			 * arguments on the stack. Depending on the type of the variable, a
			 * corresponding java-type dependant LOAD instruction must be
			 * executed.
			 */
			log.finest("Push'ing " + args + " on stack");
			int i = 1;
			for (Class<?> c : paramTypes) {
				Type paramType = Type.getType(c);
				int opcode = paramType.getOpcode(ILOAD);
				final int size = paramType.getSize();
				log.finest("Next index used by visitVarIns: " + i);
				/*
				 * declares local variables contextually to the pushing
				 * 
				 * The name is not important as its used only for debug infos,
				 * Also, there's no way to recover the original name of the
				 * variable as it was used in the original code no support in
				 * the reflection APIs. A deeper discussion might be found
				 * following links here:
				 * http://www.jroller.com/vschiavoni/entry/how_to_get_the_name
				 * Hereby, I use 'a' for every field. Any other valid character
				 * for field names could be used.
				 */
				mv.visitLocalVariable("a", paramType.toString(), null, l0, l1,
						i);
				mv.visitVarInsn(opcode, i);
				i = i + size;
			}

			mv.visitMethodInsn(INVOKEINTERFACE,
					unremotableInterfazeInternalName, method.getName(), desc);

			mv.visitInsn(Type.getReturnType(method).getOpcode(IRETURN));

			/* this should not be required because of the COMPUTE_MAX */
			mv.visitMaxs(0, 0);
			mv.visitEnd();

			log.finer("Method " + method.toGenericString() + " generated");
		}

		cw.visitEnd();
		byte[] bytecode = cw.toByteArray();
		BytecodeVerifier.verify(bytecode);

		final Class<?> definedClazz = cl.defineClass(bytecode,
				generatedClassInternalName.replace("/", "."));
		log.finest("Class: " + definedClazz + " defined by classloader " + cl);

		return definedClazz;

	}
}
